﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using Microsoft.Kinect;
using KinectUtils;

using System.IO.Ports;

namespace KinectRobot
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();
        }

        #region Kinect setup

        KinectManager myKinectMgr;
        KinectSensor myKinect;

        void setupKinect()
        {
            try
            {
                myKinectMgr = new KinectManager(
                    0, // Sensor number 0
                    ColorImageFormat.RgbResolution640x480Fps30, // Color image needed
                    DepthImageFormat.Resolution320x240Fps30,    // Depth image needed
                    true);                                      // Skeleton image needed

                myKinect = myKinectMgr.Kinect;

                myKinectMgr.DoProcessVideoFrame +=
                    new KinectManager.ProcessVideoFrame(processColorFrame);
                myKinectMgr.DoProcessSkeletons +=
                    new KinectManager.ProcessSkeletons(processSkeletonFrame);

                myKinectMgr.DoStatusChanged +=
                    new KinectManager.StatusChanged(updateStatus);

                myKinectMgr.Start();
            }
            catch
            {
                MessageBox.Show("Kinect initialise failed", "Robot Controller");
                Application.Current.Shutdown();
            }
        }

        void shutdownKinect()
        {
            if (myKinectMgr != null)
                myKinectMgr.Stop();
        }

        #endregion

        #region Kinect Video Display

        byte[] colorData = null;

        WriteableBitmap colorImage = null;
        int frameWidth;
        int frameHeight;

        private static readonly int Bgr32BytesPerPixel = (PixelFormats.Bgr32.BitsPerPixel + 7) / 8;

        void processColorFrame(ColorImageFrame colorFrame)
        {
#if SREENSHOT_PAUSE
            if (trackCount == trackLimit) return;
#endif

            frameWidth = colorFrame.Width;
            frameHeight = colorFrame.Height;
            if (colorData == null)
                colorData = new byte[colorFrame.PixelDataLength];

            colorFrame.CopyPixelDataTo(colorData);
            //Dispatcher.Invoke(new Action(() => drawColorFrame()));
        }

        void drawColorFrame()
        {
            if (colorData == null) return;

            if (colorImage == null)
            {
                this.colorImage = new WriteableBitmap(
                    frameWidth,
                    frameHeight,
                    96,  // DpiX
                    96,  // DpiY
                    PixelFormats.Bgr32,
                    null);
                kinectVideoImage.Source = colorImage;
            }

            this.colorImage.WritePixels(
                new Int32Rect(0, 0, frameWidth, frameHeight),
                this.colorData,
                frameWidth * Bgr32BytesPerPixel,
                0);
        }

        #endregion

        #region Kinect Skeleton Display

        Skeleton[] skeletons = null;
        string qualityMessage = "";

        void processSkeletonFrame(SkeletonFrame frame)
        {
            skeletons = new Skeleton[frame.SkeletonArrayLength];
            frame.CopySkeletonDataTo(skeletons);
            Dispatcher.Invoke(new Action(() => drawSkeleton()));
        }

        void drawSkeleton()
        {
            if (skeletons == null) return;

            qualityMessage = "";

            playerCanvas.Children.Clear();

            foreach (Skeleton skeleton in skeletons)
            {
                if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                {
                    if (skeleton.ClippedEdges == 0)
                    {
                        qualityMessage = "Good Quality";
                    }
                    else
                    {
                        if ((skeleton.ClippedEdges & FrameEdges.Bottom) != 0)
                            qualityMessage += "Move up ";
                        if ((skeleton.ClippedEdges & FrameEdges.Top) != 0)
                            qualityMessage += "Move down ";
                        if ((skeleton.ClippedEdges & FrameEdges.Right) != 0)
                            qualityMessage += "Move left ";
                        if ((skeleton.ClippedEdges & FrameEdges.Left) != 0)
                            qualityMessage += "Move right ";
                    }

                    // Uncomment if you want to draw the bones of the player
                    drawSkeleton(skeleton);
                }
            }
            SkeletonQualityTextBlock.Text = qualityMessage;

            // use the skeleton to control the robot
            gestureRobot(skeletons);
        }

        Brush skeletonBrush = new SolidColorBrush(Colors.Red);

        void addLine(Joint j1, Joint j2)
        {
            Line boneLine = new Line();
            boneLine.Stroke = skeletonBrush;
            boneLine.StrokeThickness = 5;

            ColorImagePoint j1P = myKinect.MapSkeletonPointToColor(j1.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X1 = j1P.X;
            boneLine.Y1 = j1P.Y;

            ColorImagePoint j2P = myKinect.MapSkeletonPointToColor(j2.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X2 = j2P.X;
            boneLine.Y2 = j2P.Y;

            playerCanvas.Children.Add(boneLine);
        }

        void drawSkeleton(Skeleton skeleton)
        {
            // Spine
            addLine(skeleton.Joints[JointType.Head], skeleton.Joints[JointType.ShoulderCenter]);
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.Spine]);

            // Left leg
            addLine(skeleton.Joints[JointType.Spine], skeleton.Joints[JointType.HipCenter]);
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipLeft]);
            addLine(skeleton.Joints[JointType.HipLeft], skeleton.Joints[JointType.KneeLeft]);
            addLine(skeleton.Joints[JointType.KneeLeft], skeleton.Joints[JointType.AnkleLeft]);
            addLine(skeleton.Joints[JointType.AnkleLeft], skeleton.Joints[JointType.FootLeft]);

            // Right leg
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipRight]);
            addLine(skeleton.Joints[JointType.HipRight], skeleton.Joints[JointType.KneeRight]);
            addLine(skeleton.Joints[JointType.KneeRight], skeleton.Joints[JointType.AnkleRight]);
            addLine(skeleton.Joints[JointType.AnkleRight], skeleton.Joints[JointType.FootRight]);

            // Left arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderLeft]);
            addLine(skeleton.Joints[JointType.ShoulderLeft], skeleton.Joints[JointType.ElbowLeft]);
            addLine(skeleton.Joints[JointType.ElbowLeft], skeleton.Joints[JointType.WristLeft]);
            addLine(skeleton.Joints[JointType.WristLeft], skeleton.Joints[JointType.HandLeft]);

            // Right arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderRight]);
            addLine(skeleton.Joints[JointType.ShoulderRight], skeleton.Joints[JointType.ElbowRight]);
            addLine(skeleton.Joints[JointType.ElbowRight], skeleton.Joints[JointType.WristRight]);
            addLine(skeleton.Joints[JointType.WristRight], skeleton.Joints[JointType.HandRight]);

        }

        #endregion

        #region Status Display

        string newStatus;

        void updateStatus(string status)
        {
            newStatus = status;
            Dispatcher.Invoke(new Action(() => updateMessage()));
        }

        void updateMessage()
        {
            KinectStatusTextBlock.Text = newStatus;
        }

        #endregion

        #region Command Decode

        // List of the touchAreas 
        List<TextBlock> touchAreas = new List<TextBlock>();

        void makeTouchArea(double x, double y, double width, double height, Brush areaBrush, string command)
        {
            // Make a new touchArea
            TextBlock touchArea = new TextBlock();

            // Set the touchArea size
            touchArea.Width = width;
            touchArea.Height = height;

            // Set the background
            touchArea.Background = areaBrush;

            // Place the touchArea on the keyboard
            touchCanvas.Children.Add(touchArea);

            // Position the touchArea
            Canvas.SetTop(touchArea, y);
            Canvas.SetBottom(touchArea, y + height);

            Canvas.SetLeft(touchArea, x);
            Canvas.SetRight(touchArea, x + width);

            // Put the command on the touchArea
            touchArea.Text = command;
            touchArea.TextAlignment = TextAlignment.Center;
            touchArea.FontSize = 15;

            // put the touchArea in the list
            touchAreas.Add(touchArea);
        }

        Brush touchBrush = new SolidColorBrush(Colors.Green);

        private void setupCommands()
        {
            double keyWidth = touchCanvas.Width / 7;

            makeTouchArea((touchCanvas.Width - keyWidth) / 2, 0, keyWidth, keyWidth, touchBrush, "Forward");
            makeTouchArea((touchCanvas.Width - keyWidth) / 2, 3 * keyWidth, keyWidth, keyWidth, touchBrush, "Back");
            makeTouchArea(0, keyWidth, keyWidth, keyWidth, touchBrush, "Left");
            makeTouchArea(touchCanvas.Width - keyWidth, keyWidth, keyWidth, keyWidth, touchBrush, "Right");
            makeTouchArea((touchCanvas.Width - keyWidth) / 2, 1.5 * keyWidth, keyWidth, keyWidth, touchBrush, "Stop");
        }

        string lastCommand = "";

        void sendCommand(int x, int y)
        {
            foreach (TextBlock area in touchAreas)
            {
                // find out if the position is in the touchArea
                if (x < Canvas.GetLeft(area)) continue;
                if (x > Canvas.GetRight(area)) continue;
                if (y < Canvas.GetTop(area)) continue;
                if (y > Canvas.GetBottom(area)) continue;

                // if we get here the position is inside the touchArea

                if (area.Text == lastCommand)
                    // don't repeat the last command
                    return;

                lastCommand = area.Text;

                sendCommand(area.Text);
            }
        }

        void jointCommand(Joint joint)
        {
            if (joint.TrackingState != JointTrackingState.Tracked)
            {
                return;
            }

            ColorImagePoint touchPoint = myKinect.MapSkeletonPointToColor(
                joint.Position, ColorImageFormat.RgbResolution640x480Fps30);

            sendCommand(touchPoint.X, touchPoint.Y);
        }

        #endregion

        #region Robot Control

        SerialPort robotControl;

        /// <summary>
        /// Sets up the serial port for your connection to the robot.
        /// You may need to change the serial port here for the one
        /// which is 
        /// </summary>
        private void setupSerial()
        {
            try
            {
                robotControl = new SerialPort();
                robotControl.PortName = "COM3";
                robotControl.BaudRate = 19200;
                robotControl.StopBits = StopBits.One;
                robotControl.Parity = Parity.None;
                robotControl.DataBits = 8;
                robotControl.ReadTimeout = 200;
                robotControl.WriteTimeout = 200;
                robotControl.DataReceived += new SerialDataReceivedEventHandler(robotControl_DataReceived);
                robotControl.Open();
            }
            catch
            {
                System.Windows.MessageBox.Show("Failed to open serial port. Please make sure the xBee device is connected.");
            }
        }

        /// <summary>
        /// Closes the serial port if it has been created
        /// </summary>
        private void CloseSerial()
        {
            if (robotControl != null)
                robotControl.Close();
        }

        void robotControl_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            Dispatcher.Invoke(new Action(() => updateResponseDisplay()));
        }

        private void updateResponseDisplay()
        {
            string text = robotControl.ReadExisting();
            robotResponseTextBox.AppendText(text);
        }

        private void updateSerialLog()
        {
            string text = robotControl.ReadExisting();

            robotResponseTextBox.AppendText(text);
        }

        private void sendCommand(string message)
        {
            message = "<" + message + ">";
            robotCommandTextBox.Text = message;
            robotControl.Write(message);
        }

        private float jointDistance(Joint first, Joint second)
        {
            float dx = first.Position.X - second.Position.X;
            float dy = first.Position.Y - second.Position.Y;
            float dz = first.Position.Z - second.Position.Z;

            return (float)Math.Sqrt((dx * dx) + (dy * dy) + (dz * dz));
        }

        private float oldDistance = 100;

        private void gestureRobot(Skeleton[] skeletons)
        {
            foreach (Skeleton skeleton in skeletons)
            {
                jointCommand(skeleton.Joints[JointType.HandRight]);
            }
        }

        #endregion

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            setupSerial();
            setupKinect();
            setupCommands();
        }

        private void Windows_Closing(object sender, EventArgs e)
        {
            if (myKinectMgr != null) myKinectMgr.Stop();
            CloseSerial();
        }
    }
}
