/*
Copyright 2010 GHI Electronics LLC
Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License. 
*/

using System;
using System.IO.Ports;
using System.Runtime.CompilerServices;
using System.Threading;
using Microsoft.SPOT.Hardware;
using GHIElectronics.NETMF.Hardware;
using GHIElectronics.NETMF.FEZ;
using GHIElectronics.NETMF.Hardware.LowLevel;

namespace GHIElectronics.NETMF.FEZ
{
    public static partial class FEZ_Components
    {
        public class LED : IDisposable
        {
            OutputPort led;
            Timer timer;
            int onTime, offTime;
            bool disposed = false;
            bool timerRunning = false;
            bool blinkState;

            [MethodImpl(MethodImplOptions.Synchronized)]
            public void Dispose()
            {
                if (!disposed)
                {
                    disposed = true;
                    timerRunning = false;
                    timer.Dispose();
                    led.Dispose();
                }
            }

            public LED(FEZ_Pin.Digital pin)
            {
                led = new OutputPort((Cpu.Pin)pin, true);
                timer = new Timer(Callback, null, Timeout.Infinite, Timeout.Infinite);
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            public void SetState(bool LED_state)
            {
                if (disposed)
                    throw new ObjectDisposedException();

                if (timerRunning)
                {
                    timer.Dispose();
                    timerRunning = false;
                }

                led.Write(LED_state);
            }

            public void TurnOn()
            {
                SetState(true);
            }

            public void ShutOff()
            {
                SetState(false);
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            public void StartBlinking(int onTime_mSec, int offTime_mSec)
            {
                if (disposed)
                    throw new ObjectDisposedException();

                this.onTime = onTime_mSec;
                this.offTime = offTime_mSec;

                timerRunning = true;
                blinkState = true;
                timer.Change(0, Timeout.Infinite);
            }

            public void StopBlinking()
            {
                StartBlinking(Timeout.Infinite, 0);
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            void Callback(object o)
            {
                if (!timerRunning)
                    return;

                if (blinkState)
                {
                    led.Write(true);
                    timer.Change(onTime, Timeout.Infinite);
                }
                else
                {
                    led.Write(false);
                    timer.Change(offTime, Timeout.Infinite);
                }

                blinkState = !blinkState;
            }
        }

        public class DistanceDetector : IDisposable
        {
            AnalogIn adc;
            float Y0 = 0;
            float X0 = 0;
            float Y1 = 0;
            float X1 = 0;
            float C = 0;

            public enum SharpSensorType : byte
            {
                GP2Y0A21YK = 0,
                GP2D120 = 1,
            }

            public void Dispose()
            {
                adc.Dispose();
            }

            public DistanceDetector(FEZ_Pin.AnalogIn pin, SharpSensorType type)
            {
                adc = new AnalogIn((AnalogIn.Pin)pin);
                adc.SetLinearScale(0, 330);
                switch (type)
                {
                    case SharpSensorType.GP2Y0A21YK:
                        Y0 = 10;
                        X0 = 315;
                        Y1 = 80;
                        X1 = 30;
                        break;

                    case SharpSensorType.GP2D120:
                        Y0 = 3;
                        X0 = 315;
                        Y1 = 30;
                        X1 = 30;
                        break;
                }
                C = (Y1 - Y0) / (1 / X1 - 1 / X0);
            }

            public float GetDistance_cm()
            {
                return C / ((float)adc.Read() + (float).001) - (C / X0) + Y0;
            }
        }

        public class Button : IDisposable
        {
            InputPort button;

            public Button(FEZ_Pin.Digital pin)
            {
                button = new InputPort((Cpu.Pin)pin, false, Port.ResistorMode.PullUp);
            }

            public Button(FEZ_Pin.Interrupt pin)
            {
                button = new InterruptPort((Cpu.Pin)pin, true, Port.ResistorMode.PullUp, Port.InterruptMode.InterruptEdgeBoth);
                button.OnInterrupt += new NativeEventHandler(OnInterrupt);
            }

            void OnInterrupt(uint data1, uint data2, DateTime time)
            {
                ButtonState state = (data2 == 0) ? ButtonState.Pressed : ButtonState.NotPressed;
                ButtonPressEvent((FEZ_Pin.Interrupt)data1, state);
            }

            public void Dispose()
            {
                button.Dispose();
            }

            public enum ButtonState : byte
            {
                NotPressed = 0,
                Pressed = 1,
            }

            public ButtonState GetState()
            {
                return (button.Read() == false) ? ButtonState.Pressed : ButtonState.NotPressed;
            }

            public delegate void ButtonPressEventHandler(FEZ_Pin.Interrupt pin, ButtonState state);
            public event ButtonPressEventHandler ButtonPressEvent = delegate { };
        }
        public class Fez_Xbee
        {
            static void RemapCOM4to_TXAn2_RXAn3(SerialPort ser)
            {
                // call this function **after** you open COM4 port
                if (ser.PortName != "COM4" || ser.IsOpen == false)
                    throw new Exception("Only use COM4 and make sure it is open");
                // remap COM4 RX (in) pin from P4.29/DIO17 to P0.26 (that is An3)
                // remap COM4 TX (out) pin from P4.28/DIO13 to P0.25 (that is An2)
                Register PINSEL9 = new Register(0xE002C024);
                PINSEL9.Write(0);// COM4 is now disconnected from P4.28 and P4.29
                Register PINSEL1 = new Register(0xE002C004);
                PINSEL1.SetBits(0xf << 18);// COM4 is now connected to An3 and An4
            }

            static SerialPort xbeePort = null;

            public static SerialPort GetPort(bool usingRobot)
            {
                if (xbeePort == null)
                {
                    if (usingRobot)
                    {
                        xbeePort = new SerialPort("COM4", 19200, Parity.None, 8, StopBits.One);
                        xbeePort.Open();
                        RemapCOM4to_TXAn2_RXAn3(xbeePort);
                    }
                    else
                    {
                        xbeePort = new SerialPort("COM2", 19200, Parity.None, 8, StopBits.One);
                        xbeePort.Open();
                    }
                }
                return xbeePort;
            }
        }

    }

    static class FEZMini_Robot
    {
        static OutputCompare oc = new OutputCompare((Cpu.Pin)FEZ_Pin.Digital.An0, true, 5);
        static uint[] timings = new uint[2];

        static PWM _pwm1 = new PWM((PWM.Pin)FEZ_Pin.PWM.Di5);
        static OutputPort _dir1 = new OutputPort((Cpu.Pin)FEZ_Pin.Digital.Di3, false);

        static PWM _pwm2 = new PWM((PWM.Pin)FEZ_Pin.PWM.Di6);
        static OutputPort _dir2 = new OutputPort((Cpu.Pin)FEZ_Pin.Digital.Di9, false);

        static sbyte _last_speed_left = 0, _last_speed_right = 0;


        static public int PIEZO_RESONATE_FREQUENCY
        {
            get
            {
                return 8200;
            }
        }

        static public void Beep(int freq, int duration_mSec)
        {
            if (freq > 10000)
                throw new Exception("Frequency is too high");

            timings[0] = timings[1] = (uint)((long)1000000 / (freq / 2));
            oc.Set(true, timings, 0, 2, true);
            Thread.Sleep(duration_mSec);
            oc.Set(false);

        }

        static public void Move(sbyte speed_left, sbyte speed_right)
        {
            if (speed_left > 100 || speed_left < -100 || speed_right > 100 || speed_right < -100)
                new ArgumentException();

            _last_speed_left = speed_left;
            _last_speed_right = speed_right;


            if (speed_left < 0)
            {
                _dir1.Write(true);
                _pwm1.Set(1000, (byte)(100 - Math.Abs(speed_left)));
            }
            else
            {
                _dir1.Write(false);
                _pwm1.Set(1000, (byte)speed_left);
            }

            ////////////////////////////

            if (speed_right < 0)
            {
                _dir2.Write(true);
                _pwm2.Set(1000, (byte)(100 - Math.Abs(speed_right)));

            }
            else
            {
                _dir2.Write(false);
                _pwm2.Set(1000, (byte)speed_right);
            }
        }

        static public void MoveRamp(sbyte speed_left, sbyte speed_right, byte ramping_delay)
        {
            sbyte temp_speed_left = _last_speed_left;
            sbyte temp_speed_right = _last_speed_right;

            while ((speed_left != temp_speed_left) ||
                    (speed_right != temp_speed_right))
            {
                if (temp_speed_left > speed_left)
                    temp_speed_left--;
                if (temp_speed_left < speed_left)
                    temp_speed_left++;

                if (temp_speed_right > speed_right)
                    temp_speed_right--;
                if (temp_speed_right < speed_right)
                    temp_speed_right++;

                Move(temp_speed_left, temp_speed_right);
                Thread.Sleep(ramping_delay);
            }
        }
    }
}

